/***********************************************************************
* Copyright (c) 2005 Jan Demter                                   *
*                                                                     *
* This program is free software; you can redistribute it and/or       *
* modify it under the terms of the GNU General Public License         *
* as published by the Free Software Foundation; either version 2      *
* of the License, or (at your option) any later version.              *
*                                                                     *
* This program is distributed in the hope that it will be useful,     *
* but WITHOUT ANY WARRANTY; without even the implied warranty of      *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the       *
* GNU General Public License for more details.                        *
*                                                                     *
* You should have received a copy of the GNU General Public License   *
* along with this program; if not, write to the Free Software         *
* Foundation, Inc., 59 Temple Place - Suite 330, Boston,              *
* MA  02111-1307, USA.                                                *
*                                                                     *
* ------------------------------------------------------------------- *
* project : GONE NSLP                                                 *
* author  : Jan Demter <demter@math.uni-goettingen.de>           * 
***********************************************************************/

#include <stdexcept>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <err.h>
#include "GONEServer.h"
#include "GONEClient.h"
#include "GONE.h"
#include "objects/GONEHeader.h"
#include "../../gist/GistConfiguration.h"

GONEServer::GONEServer(bool be_verbose)
{
	verbose = be_verbose;	
	
	// read ./nsis.conf to get our local ip addresses
	gistconfig.readConfigFromFile("./nsis.conf");
	my_addr = gistconfig.general_localIPv4s[0].v4;
	
	used=false;
	
	// open ./GONEdata to store arriving data
	if(0 > (datafile = open("./GONEdata", O_CREAT | O_RDWR, S_IRWXU))){
	//if(0 > (datafile = open("/dev/null", O_RDWR))){
		
		std::cerr<<"Failed to open ./GONEdata\n";
		exit(23);
	}
	
	// open /dev/urandom, used to generate capability tokens
	if(0 > (urandom = open("/dev/urandom", O_RDONLY))){
		
		std::cerr<<"Failed to open /dev/urandom\n";
		exit(23);
	}
	
	capabilitytable = new HashTable<GONEConnectionCapability, GONEConnectionCapability>();
	connections = new HashTable<GONEConnection, GONEConnection>();
	client_cache_full = false;
	
	run();
}

void GONEServer::run()
{
	nslpSock = createServerSocket();
	clientSockCount = 0;
	srand(time(NULL));
	
	clientCallback = new NslpApiSocketCallback<GONEServer>();
	clientCallback->SetCallback(this, &GONEServer::handleClientSocket);
	
	serverCallback = new NslpApiSocketCallback<GONEServer>();
	serverCallback->SetCallback(this, &GONEServer::acceptClient);
	
	nslpApiCallback = new NslpApiCallback<GONEServer>();
	nslpApiCallback->SetCallback(this,
								 &GONEServer::handleRecvMessage,
								 &GONEServer::handleMessageStatus,
								 &GONEServer::handleNetworkNotification);
	
	api = new NslpApi();
	api->register_nslpid(GONE_NSLP_ID, true);
	api->connectToGist(nslpApiCallback);
	api->addSocket(nslpSock, NULL, serverCallback);
	
	while(true) {
		api->waitForSignal();
	}
}

GONEServer::~GONEServer()
{
	printf("GONEServer::~GONEServer\n");
	
	/*	delete api; TODO: segfault
	delete nslpApiCallback;
	delete clientCallback;
	delete serverCallback;*/
	
	unlink(NSLP_PATH_TO_UNIX_SOCKET); 
}

void GONEServer::newSession(unsigned char sid[16])
{
	printf("new sid: ");
	for(unsigned int i=0; i<16; i++) {
		sid[i] = (unsigned char)rand();
		printf("%.2x", sid[i]);
	}
	printf("\n");
}

void GONEServer::printSession(unsigned char sid[16])
{
	printf("sid: ");
	for(unsigned int i=0; i<16; i++) {
		printf("%.2x", sid[i]);
	}
	printf("\n");
}

void GONEServer::handleClientSocket(int sockfd, void * sockData){
	
	GONE_VERBOSE(std::cerr<<"GONEServer::handleClientSocket got called\n";);

	if(client_cache_full){
		
		return;

	}
	
	char buffer[16]; // will hold command from client
	memset(buffer, 0, sizeof(buffer));
	
	int res = recv(sockfd, buffer, sizeof(buffer), 0);
	if(res < 0){
		std::cerr<<"recv() went wrong\n";
		closeSocket(sockfd);
		return;
	}
	
	else if (res < (int) sizeof(buffer) && res > 0) { // FIXME?
		std::cerr<<"GONEServer::handleClientSocket: short read "<< res << std::endl;
		closeSocket(sockfd);
		return;
	}
	
	else if(res == 0){
		closeSocket(sockfd);
		return;
	}
	
	unsigned int command = ntohl(*(unsigned int *)buffer);
	
	if(command == GONE_CLIENT_SENDMESSAGE){
		
		GONEHeader sendmsg = GONEHeader((unsigned char *)buffer, sizeof(buffer)); 
		
		GONE_VERBOSE(std::cerr<<"GONE_CLIENT_SENDMESSAGE empfangen, Daten: "<< sendmsg.length() << " byte\n";)
		
		msg_routing_info_path_coupled_t pc_mri;
		pc_mri.msgRoutingMethod = GIST_DEFAULT_MRM;
		pc_mri.ipVersion = 4;
		pc_mri.srcAddr.v4.addr = my_addr;
		pc_mri.destAddr.v4.addr = sendmsg.address();
		pc_mri.srcPrefix = 32;
		pc_mri.destPrefix = 32;
		pc_mri.d = DOWNSTREAM;
		pc_mri.f = 0;
		pc_mri.flowLabel = 0;
		pc_mri.destPort = 0;
		pc_mri.srcPort = 0;
		pc_mri.spi = 0;
		pc_mri.s = 0;
		pc_mri.a = 0;
		pc_mri.b = 0;
		
		unsigned char sid[16] = {0xAA, 0xFF, 0xFF, 0xEE, 0xAA, 0xFF, 0xFF, 0xEE,0xAA, 0xFF, 0xFF, 0xEE,0xAA, 0xFF, 0xFF, 0xEE} ; // TODO...
		
		char * buffer2 = (char *)malloc(sendmsg.getLength()+sendmsg.length());
		
		// does not work for large stuff
		int res = recv(sockfd, buffer2+sendmsg.getLength(), sendmsg.length(), 0);
		if (res < (int)sendmsg.length()) { //especially this comparison...
			
			std::cerr<<"GONE_CLIENT_SENDMESSAGE: message too large for this messy implementation\n";
			closeSocket(sockfd);
			free(buffer2);
			return;
		}
		
		// find matching GONEConnection, set one up if none
		GONEConnection this_connection_temp = GONEConnection(my_addr, sendmsg.address(), 23);
		
		GONEConnection * this_connection = connections->get(&this_connection_temp); // our connection or 0
				
		if(this_connection != NULL){ // there is already a matching connection, try to send the stuff out
			GONE_VERBOSE(std::cerr<<"matching connection found\n";)
			
			if(!this_connection->ready){ // connection not ready yet, write to buffer
				GONE_VERBOSE(std::cerr<<"connection not ready yet, buffering.\n";)
				
				if( (this_connection->buffer_length+sendmsg.length()) < MAX_BUF_LEN){
					
					memcpy(this_connection->buffer+this_connection->buffer_length, buffer2+sendmsg.getLength(),
						   sendmsg.length());
					
					this_connection->buffer_length+=sendmsg.length();
					free(buffer2);
					client_cache_full = true;
					return;
				}
				else{
					std::cerr<<"buffer full, dropping data.\n";
					// TODO: this should somehow be announced back to the app...
					client_cache_full = true;
					free(buffer2);
					return;
				}
				
			}
			
			// get capability
			if(this_connection->next_cap < GONE_CAPTOKENS)
				sendmsg.captokenlist(this_connection->capabilities[this_connection->next_cap++]);
			else{
				GONE_VERBOSE(std::cerr<<"We're out of capability tokens, buffering...\n";)				
				// connection not ready anymore:
				this_connection->ready = false;
				client_cache_full = true;
				
				// Buffer
				if( (this_connection->buffer_length+sendmsg.length()) < MAX_BUF_LEN){
					
					memcpy(this_connection->buffer+this_connection->buffer_length, buffer2+sendmsg.getLength(),
						   sendmsg.length());
					
					this_connection->buffer_length+=sendmsg.length();
					
					free(buffer2);
				}
				else{
					std::cerr<<"buffer full while out of captokens, dropping data.\n";
					// TODO: this should somehow be announced back to the app...
					free(buffer2);
				}
				
				// data is buffered (or not) now, get new tokens:
				
				GONE_VERBOSE(std::cerr<<"Trying to get new capability tokens\n";)
				GONEHeader message_header = GONEHeader(GONE_SETUPREQ, 0,
													   sendmsg.address(), 0);
				
				msg_routing_info_path_coupled_t pc_mri;
				pc_mri.msgRoutingMethod = GIST_DEFAULT_MRM;
				pc_mri.ipVersion = 4;
				pc_mri.srcAddr.v4.addr = my_addr;
				pc_mri.destAddr.v4.addr = sendmsg.address();
				pc_mri.srcPrefix = 32;
				pc_mri.destPrefix = 32;
				pc_mri.d = DOWNSTREAM;
				pc_mri.f = 0;
				pc_mri.flowLabel = 0;
				pc_mri.destPort = 0;
				pc_mri.srcPort = 0;
				pc_mri.spi = 0;
				pc_mri.s = 0;
				pc_mri.a = 0;
				pc_mri.b = 0;
				
				unsigned char sid[16] = {0xAA, 0xFF, 0xFF, 0xEE, 0xAA, 0xFF, 0xFF, 0xEE,0xAA, 0xFF, 0xFF, 0xEE,0xAA, 0xFF, 0xFF, 0xEE};
				api->SendMessage((char *)message_header.getBuffer(),
								 //sizeof(buffer), NULL, GONE_NSLP_ID,
								 message_header.getLength(), NULL, GONE_NSLP_ID,
								 (unsigned char *) sid,
								 (msg_routing_info_base_t *)&pc_mri, NULL, NULL, true, false, false,
								 2000, 10);

				free(buffer2);
				return;
			}
			
			// fiddle it together
			memcpy(buffer2, buffer, sendmsg.getLength());
						
			// and spread the word
			api->SendMessage((char *)buffer2, sendmsg.getLength()+sendmsg.length(),
							 NULL, GONE_NSLP_ID, (unsigned char *) sid, (msg_routing_info_base_t *)&pc_mri,
							 NULL, NULL, true, false, false,
							 2000, 10);
			
			this_connection->used = true;
			this_connection->timestamp = time(NULL);
			
			//clean up
			free(buffer2);
			return;
		} // end if(connections->get(this_connection) != NULL)
		
		// no matching connection, establish a new one
		else{
			GONE_VERBOSE(std::cerr<<"Trying to establish a new connection\n";);

			client_cache_full = true;

			//print timestamp
			/*
			struct timeval diezeit;
			gettimeofday(&diezeit, NULL);
			//std::cout<< " SETUP START: " << diezeit.tv_sec << "." << std::setw(6) << std::setfill('0') << diezeit.tv_usec <<":";
			printf("SETUP START: %i.%06d:", diezeit.tv_sec, diezeit.tv_usec);
			*/
									
			GONEHeader message_header = GONEHeader(GONE_SETUPREQ, 0,
												   sendmsg.address(), 0);
			
			GONEConnection * new_connection = new GONEConnection(my_addr, sendmsg.address(), 23);
			
			if( (new_connection->buffer_length+sendmsg.length()) < MAX_BUF_LEN ){
				memcpy(new_connection->buffer+new_connection->buffer_length, buffer2+sendmsg.getLength(),
					   sendmsg.length());
				
				new_connection->buffer_length+=sendmsg.length();
			}
			else{
				// FIXME...
				std::cerr<<"buffer too small, dropping data\n";
			}
			
			connections->put(new_connection, new_connection);
			
			msg_routing_info_path_coupled_t pc_mri;
			pc_mri.msgRoutingMethod = GIST_DEFAULT_MRM;
			pc_mri.ipVersion = 4;
			pc_mri.srcAddr.v4.addr = my_addr;
			pc_mri.destAddr.v4.addr = sendmsg.address();
			pc_mri.srcPrefix = 32;
			pc_mri.destPrefix = 32;
			pc_mri.d = DOWNSTREAM;
			pc_mri.f = 0;
			pc_mri.flowLabel = 0;
			pc_mri.destPort = 0;
			pc_mri.srcPort = 0;
			pc_mri.spi = 0;
			pc_mri.s = 0;
			pc_mri.a = 0;
			pc_mri.b = 0;
			
			unsigned char sid[16] = {0xAA, 0xFF, 0xFF, 0xEE, 0xAA, 0xFF, 0xFF, 0xEE,0xAA, 0xFF, 0xFF, 0xEE,0xAA, 0xFF, 0xFF, 0xEE};
			api->SendMessage((char *)message_header.getBuffer(),
							 //sizeof(buffer), NULL, GONE_NSLP_ID,
							 message_header.getLength(), NULL, GONE_NSLP_ID,
							 (unsigned char *) sid,
							 (msg_routing_info_base_t *)&pc_mri, NULL, NULL, true, false, false,
							 2000, 10);
			
		}
		
	} // if(command == GONE_CLIENT_SENDMESSAGE)
	
	else if(command == GONE_CLIENT_FIN){
		
		GONEHeader clientmsg = GONEHeader((unsigned char *)buffer, sizeof(buffer));
		
		//find matching connection
		GONEConnection this_connection_temp = GONEConnection(my_addr, clientmsg.address(), 23);
		
		GONEConnection * this_connection = connections->get(&this_connection_temp);
		
		if(this_connection != NULL){
		GONEHeader finmsg = GONEHeader(GONE_FIN, GONE_CAPTOKENS*sizeof(unsigned int),
											   clientmsg.address(), GONE_CAPTOKENS);
		
		//fiddle header and captokens together
		unsigned int length = finmsg.getLength()+GONE_CAPTOKENS*sizeof(unsigned int);
		char * sendbuffer = new char[length];
		
		memcpy(sendbuffer, finmsg.getBuffer(), finmsg.getLength());
		memcpy(sendbuffer+finmsg.getLength(), this_connection->capabilities, GONE_CAPTOKENS*sizeof(unsigned int));
		
		//send fin out:
		msg_routing_info_path_coupled_t pc_mri;
		pc_mri.msgRoutingMethod = GIST_DEFAULT_MRM;
		pc_mri.ipVersion = 4;
		pc_mri.srcAddr.v4.addr = my_addr;
		pc_mri.destAddr.v4.addr = clientmsg.address();
		pc_mri.srcPrefix = 32;
		pc_mri.destPrefix = 32;
		pc_mri.d = DOWNSTREAM;
		pc_mri.f = 0;
		pc_mri.flowLabel = 0;
		pc_mri.destPort = 0;
		pc_mri.srcPort = 0;
		pc_mri.spi = 0;
		pc_mri.s = 0;
		pc_mri.a = 0;
		pc_mri.b = 0;
		
		unsigned char sid[16] = {0xAA, 0xFF, 0xFF, 0xEE, 0xAA, 0xFF, 0xFF, 0xEE,0xAA, 0xFF, 0xFF, 0xEE,0xAA, 0xFF, 0xFF, 0xEE};	//FIXME
		
		// spread the word
		api->SendMessage(sendbuffer, length, NULL,
						 GONE_NSLP_ID, (unsigned char *) sid,
						 (msg_routing_info_base_t *)&pc_mri, NULL, NULL, true, false, false,
						 2000, 10);

		connections->remove(this_connection);
		delete this_connection;
		return;
	
		}
		else{
			std::cerr<<"No matching connection for FIN found\n";
			return;
		}
		
		
	}
	else{
		std::cerr<<"Unknown command\n";
		return;
	}
}

void GONEServer::handleRecvMessage(api_recv_msg * s){
	// handle incoming messages here (GIST communication)
	
	GONE_VERBOSE(printf("GONEServer::handleRecvMessage() got a message! (%u bytes)\n", s->nd_size);)
	
	// check whether we're the destination for this stuff
	//if( gistconfig.isLocalIPaddr(4, (unsigned char *)&(s->mri->destAddr.v4.addr)) ){
	
	GONE_VERBOSE(printf("Surprise, surprise: We got a message. Just for us! Rejoice!\n");)
	
	unsigned int type = ntohl(*(unsigned int *)s->nslp_data);
	
	if(type == GONE_DATA){
		GONE_VERBOSE(std::cerr<<"received GONE_DATA";)
		
		GONEHeader sentmsg = GONEHeader
			((unsigned char *)s->nslp_data, 16); //FIXME
		
		GONEConnectionCapability temp_cap = GONEConnectionCapability(s->mri->srcAddr.v4.addr,
																	 s->mri->destAddr.v4.addr,
																	 sentmsg.captokenlist()
																	 );
		
		// Is the message is for this host?
		if( gistconfig.isLocalIPaddr(4, (unsigned char *)&(s->mri->destAddr.v4.addr)) ){
			
			GONE_VERBOSE(std::cerr<<" for us\n";)

			//find matching connection
			GONEConnection temp_con = GONEConnection(s->mri->srcAddr.v4.addr, s->mri->destAddr.v4.addr,
																 23);
			GONEConnection * this_con = connections->get(&temp_con);

			if(this_con == NULL){
				std::cerr<<"\tNo matching connection found\n";
				return;
			}
			
			GONEConnectionCapability * the_cap;
			
			if((the_cap = capabilitytable->get(&temp_cap)) != NULL){
				
				char * daten = s->nslp_data+sentmsg.getLength();
				
				//daten[s->nd_size-1] = '\0';
				//printf("Sent message is: \"%s\"\n", daten);
				
				//FIXME: this is just for testing.
				ssize_t written = write(datafile, daten, sentmsg.length());
				if(written < 0)
					err(1,"write(): %s, %i", __FILE__, __LINE__);
				GONE_VERBOSE(std::cerr<<"Wrote " << written << " bytes\n";)
				
								
				if(!this_con->used){
					this_con->used=true;
					struct timeval diezeit;
					gettimeofday(&diezeit, NULL);
					fprintf(stderr, "TRANSFER START: %i.%06d:", diezeit.tv_sec, diezeit.tv_usec);
				}
				
				//delete capability
				capabilitytable->remove(the_cap);
				delete the_cap;
				this_con->next_cap++;
			}
			else
				std::cerr<<"Message contained wrong capability token\n";
		}
		
		// Message is to be forwarded, check capabilitiy token
		else{
			GONE_VERBOSE(std::cerr<<" for someone else\n";)
			
			GONEConnectionCapability temp_cap = GONEConnectionCapability(s->mri->srcAddr.v4.addr,
																		 s->mri->destAddr.v4.addr,
																		 sentmsg.captokenlist()
																		 );
			GONEConnectionCapability * the_cap;
			
			if((the_cap = capabilitytable->get(&temp_cap)) != NULL){
				
				api->SendMessage(s->nslp_data, s->nd_size, NULL, GONE_NSLP_ID,
								 (unsigned char *)s->sid,
								 (msg_routing_info_base_t *)s->mri, NULL, NULL, true, false, false,
								 2000, 10);
				
				//delete capability
				capabilitytable->remove(the_cap);
				delete the_cap;
				return;
			}
			else
				std::cerr<<"Message to be forwarded contained wrong capability token:" << sentmsg.captokenlist() << "\n";
		}
	} // if(type == GONE_DATA)
	
	else if(type == GONE_SETUPREQ){
		GONE_VERBOSE(std::cerr<<"Received GONE_SETUPREQ";)
		
		// Is the setup request for this host?
		if( gistconfig.isLocalIPaddr(4, (unsigned char *)&(s->mri->destAddr.v4.addr)) ){
			
			
			GONE_VERBOSE(std::cerr<<" for us\n";);
			
			//find matching connection, if one
			GONEConnection temp_con = GONEConnection(s->mri->srcAddr.v4.addr, s->mri->destAddr.v4.addr,
													 23);
			GONEConnection * this_con = connections->get(&temp_con);
			
			if(this_con == NULL){
				GONE_VERBOSE(std::cerr<<"\tSetting up a new connection\n";);
				this_con = new GONEConnection(s->mri->srcAddr.v4.addr, s->mri->destAddr.v4.addr,
																	 23);
				connections->put(this_con, this_con);
			}			
			
			//GONEHeader * setupreq = new GONEHeader((unsigned char *)s->nslp_data, 16);
			
			GONEHeader setupresponse = GONEHeader(GONE_SETUPRES, GONE_CAPTOKENS*sizeof(unsigned int),
														 s->mri->srcAddr.v4.addr, GONE_CAPTOKENS);
			
			unsigned int sendbuffer_length = setupresponse.getLength()+setupresponse.length();
			
			char * sendbuffer = new char[sendbuffer_length];
			
			//TODO: capability tokens should of course not be static..., FIXME: byteorder
			//unsigned int captokenarray[16] = {1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16};
			
			// read GONE_CAPTOKENS*sizeof(unsigned int) more or less random data in
			// FIXME: byteorder
			unsigned int captokenarray[GONE_CAPTOKENS];
			if( GONE_CAPTOKENS*sizeof(unsigned int) != read(urandom, &captokenarray, GONE_CAPTOKENS*sizeof(unsigned int))){
			
				std::cerr<<"Error while reading from /dev/urandom\n";
			}
				
			memcpy(this_con->capabilities, captokenarray, 16*4);

			this_con->next_cap = 0;			
			
			for(int i = 0; i<GONE_CAPTOKENS; i++){
				GONEConnectionCapability * cap = new GONEConnectionCapability(s->mri->srcAddr.v4.addr,
																			  s->mri->destAddr.v4.addr,
																			  captokenarray[i]);
				capabilitytable->put(cap, cap);
				//std::cerr<<"captokenarray["<< i <<"] = " << captokenarray[i] << std::endl;
			}
			
			//fiddle captokenarray and GONEHeader together, TODO: check for errors
			memcpy(sendbuffer, setupresponse.getBuffer(), setupresponse.getLength());
			memcpy(sendbuffer+setupresponse.getLength(), captokenarray, GONE_CAPTOKENS*sizeof(unsigned int));
			
			msg_routing_info_path_coupled_t pc_mri;
			pc_mri.msgRoutingMethod = GIST_DEFAULT_MRM;
			pc_mri.ipVersion = 4;
			pc_mri.srcAddr.v4.addr = my_addr;
			pc_mri.destAddr.v4.addr = s->mri->srcAddr.v4.addr;
			pc_mri.srcPrefix = 32;
			pc_mri.destPrefix = 32;
			pc_mri.d = DOWNSTREAM;
			pc_mri.f = 0;
			pc_mri.flowLabel = 0;
			pc_mri.destPort = 0;
			pc_mri.srcPort = 0;
			pc_mri.spi = 0;
			pc_mri.s = 0;
			pc_mri.a = 0;
			pc_mri.b = 0;
			
			unsigned char sid[16] = {0xAA, 0xFF, 0xFF, 0xEE, 0xAA, 0xFF, 0xFF, 0xEE,0xAA, 0xFF, 0xFF, 0xEE,0xAA, 0xFF, 0xFF, 0xEE};	//FIXME
			
			// spread the word
			api->SendMessage(sendbuffer, setupresponse.getLength()+setupresponse.captokenlist()*4, NULL,
							 GONE_NSLP_ID, (unsigned char *) sid,
							 (msg_routing_info_base_t *)&pc_mri, NULL, NULL, true, false, false,
							 2000, 10);
			//std::cerr<<"sent:\n setupresponse->length(): " << setupresponse.length() <<
			//	"\nsetupresponse->captokenlist(): " << setupresponse.captokenlist() << "\n";

			this_con->ready=true;

			delete sendbuffer;
		}
		else{
			//Pass setup request on
			//TODO: do some rate limiting on this to guard against DoS
			
			GONE_VERBOSE(std::cerr << "Passing on a GONE setup request\n";)
			
			/*
			 msg_routing_info_path_coupled_t pc_mri;
			 pc_mri.msgRoutingMethod = GIST_DEFAULT_MRM;
			 pc_mri.ipVersion = 4;
			 pc_mri.srcAddr.v4.addr = s->mri->srcAddr.v4.addr;
			 pc_mri.destAddr.v4.addr = s->mri->destAddr.v4.addr;
			 pc_mri.srcPrefix = 32;
			 pc_mri.destPrefix = 32;
			 pc_mri.d = DOWNSTREAM;
			 pc_mri.flowLabel = 0;
			 pc_mri.destPort = 0;
			 pc_mri.srcPort = 0;
			 pc_mri.spi = 0;
			 */
			
			api->SendMessage(s->nslp_data, s->nd_size, NULL, GONE_NSLP_ID,
							 (unsigned char *)s->sid,
							 (msg_routing_info_base_t *)s->mri, NULL, NULL, true, false, false,
							 2000, 10);
			
		}
		
	} // if(type == GONE_SETUPREQ)

	else if(type == GONE_SETUPRES){
		GONE_VERBOSE(std::cerr<<"Got GONE_SETUPRES";)
		
		// Is the setup request for this host?
		if( gistconfig.isLocalIPaddr(4, (unsigned char *)&(s->mri->destAddr.v4.addr)) ){
			
			GONEHeader setupresponse = GONEHeader((unsigned char *)s->nslp_data, 16);
			
			GONE_VERBOSE(std::cerr<<" for us, length is " << s->nd_size << "\n";)
			//std::cerr<<"setupresponse->captokenlist() = " << setupresponse->captokenlist() << std::endl;
			//std::cerr<<"setupresponse->length() = " << setupresponse->length() << std::endl;
			
			
			//find matching connection
			
			GONEConnection * this_connection_temp = new GONEConnection(s->mri->destAddr.v4.addr,
																	   s->mri->srcAddr.v4.addr, 23);
			
			GONEConnection * this_connection = connections->get(this_connection_temp);
			delete this_connection_temp;
			
			//check if we found one
			if(this_connection == NULL){
				std::cerr<<"No matching connection for setup response found\n";
				return;
			}
			
			//save capability tokens
			if(s->nd_size-setupresponse.captokenlist()*4 == setupresponse.getLength()){
				memcpy(this_connection->capabilities, s->nslp_data+setupresponse.getLength(),
					   GONE_CAPTOKENS*sizeof(unsigned int));
				this_connection->next_cap = 0;
				
			}
			else{
				std::cerr<<"Mismatch between captokenlist length and length of msg ("<< s->nd_size-setupresponse.captokenlist()*4 << "\n";
				/*
				std::cerr<<"setupresponse->captokenlist() = " << setupresponse->captokenlist() << std::endl;
				in_addr dings;
				dings.s_addr = setupresponse->address();
				std::cerr<<"setupresponse->address() = " << inet_ntoa(dings) << std::endl;
				std::cerr<<"setupresponse->length() = " << setupresponse->captokenlist() << std::endl;
				 */
				
				return;
			}

			// print timestamp
			/*
			struct timeval diezeit;
			gettimeofday(&diezeit, NULL);
			//std::cout<< diezeit.tv_sec << "." std::setw(6) << std::setfill('0') << diezeit.tv_usec <<" SETUP ENDE\n";
			fprintf(stderr, "%i.%06d SETUP ENDE\n", diezeit.tv_sec, diezeit.tv_usec);
			*/
												
			
			//send out buffered data
			
			GONEHeader message_header = GONEHeader(GONE_DATA, this_connection->buffer_length,
												   s->mri->srcAddr.v4.addr,
												   this_connection->capabilities[this_connection->next_cap++]);
			
			//fiddle stuff together
			char * sendbuffer = (char *)malloc(message_header.getLength()+this_connection->buffer_length);
			memcpy(sendbuffer, message_header.getBuffer(), message_header.getLength());
			memcpy(sendbuffer+message_header.getLength(), this_connection->buffer, this_connection->buffer_length);
			
			msg_routing_info_path_coupled_t pc_mri;
			pc_mri.msgRoutingMethod = GIST_DEFAULT_MRM;
			pc_mri.ipVersion = 4;
			pc_mri.srcAddr.v4.addr = my_addr;
			pc_mri.destAddr.v4.addr = s->mri->srcAddr.v4.addr;
			pc_mri.srcPrefix = 32;
			pc_mri.destPrefix = 32;
			pc_mri.d = DOWNSTREAM;
			pc_mri.f = 0;
			pc_mri.flowLabel = 0;
			pc_mri.destPort = 0;
			pc_mri.srcPort = 0;
			pc_mri.spi = 0;
			pc_mri.s = 0; // What is this?
			pc_mri.a = 0;
			pc_mri.b = 0;
			
			unsigned char sid[16] = {0xAA, 0xFF, 0xFF, 0xEE, 0xAA, 0xFF, 0xFF, 0xEE,0xAA, 0xFF, 0xFF, 0xEE,0xAA, 0xFF, 0xFF, 0xEE};	//FIXME
			
			
			
			// and spread the word
			api->SendMessage(sendbuffer, message_header.getLength()+this_connection->buffer_length, NULL,
							 GONE_NSLP_ID, (unsigned char *) sid,
							 (msg_routing_info_base_t *)&pc_mri, NULL, NULL, true, false, false,
							 2000, 10);

			//connection should be ready now
			this_connection->ready = true;
			this_connection->buffer_length = 0;
			client_cache_full = false;

			free(sendbuffer);
		}
		else{
			//Pass setup response on and save capabilities
			//TODO: do some rate limiting on this to guard against DoS
			
			GONEHeader setupresponse = GONEHeader((unsigned char *)s->nslp_data, 16);
			
			GONE_VERBOSE(std::cerr<<" for someone else, length is " << s->nd_size << "\n";)
			//std::cerr<<"setupresponse->captokenlist() = " << setupresponse->captokenlist() << std::endl;
			//std::cerr<<"setupresponse->length() = " << setupresponse->length() << std::endl;
			
			//std::cerr << "Reading capability tokens\n";
			
			if(s->nd_size-setupresponse.captokenlist()*4 == setupresponse.getLength()){
				
				//unsigned int * captokenarray = (unsigned int *)s->nslp_data + setupresponse->getLength();
				
				unsigned int captokenarray[16];
				
				memcpy(captokenarray, s->nslp_data + setupresponse.getLength(), GONE_CAPTOKENS*sizeof(unsigned int));
				
				for(int i = 0; i<GONE_CAPTOKENS; i++){
					GONEConnectionCapability * cap = new GONEConnectionCapability(s->mri->destAddr.v4.addr,
																				  s->mri->srcAddr.v4.addr,
																				  captokenarray[i]);
					capabilitytable->put(cap, cap);
					//std::cerr<<"captokenarray["<< i <<"] = " << captokenarray[i] << std::endl;
				}
				
			}
			
			/*
			 msg_routing_info_path_coupled_t pc_mri;
			 pc_mri.msgRoutingMethod = GIST_DEFAULT_MRM;
			 pc_mri.ipVersion = 4;
			 pc_mri.srcAddr.v4.addr = s->mri->srcAddr.v4.addr;
			 pc_mri.destAddr.v4.addr = s->mri->destAddr.v4.addr;
			 pc_mri.srcPrefix = 32;
			 pc_mri.destPrefix = 32;
			 pc_mri.d = DOWNSTREAM;
			 pc_mri.flowLabel = 0;
			 pc_mri.destPort = 0;
			 pc_mri.srcPort = 0;
			 pc_mri.spi = 0;
			 */
			
			api->SendMessage(s->nslp_data, s->nd_size, NULL, GONE_NSLP_ID,
							 (unsigned char *)s->sid,
							 (msg_routing_info_base_t *)s->mri, NULL, NULL, true, false, false,
							 2000, 10);
			
		}
		
	} // if(type == GONE_SETUPRES)
	
	else if(type == GONE_FIN){
		GONE_VERBOSE(std::cerr<<"Got GONE_FIN";);
		
		GONEHeader finmsg = GONEHeader((unsigned char *)s->nslp_data, 16);
		
		// Is the setup request for this host?
		if( gistconfig.isLocalIPaddr(4, (unsigned char *)&(s->mri->destAddr.v4.addr)) ){
			
			GONE_VERBOSE(std::cerr<<" for us, length is " << s->nd_size << "\n";);
			
			GONEConnection * this_connection_temp = new GONEConnection(s->mri->srcAddr.v4.addr,
																	   s->mri->destAddr.v4.addr, 23);
			
			GONEConnection * this_connection = connections->get(this_connection_temp);
			delete this_connection_temp;
			
			//check if we found one
			if(this_connection == NULL){
				std::cerr<<"No matching connection for fin msg found\n";
				return;
			}
			
			//FIXME: delete remaining cap tokens, connection
			//delete remaining capabilities
			if(s->nd_size-finmsg.captokenlist()*4 == finmsg.getLength()){
				
				unsigned int captokenarray[16];
				
				memcpy(captokenarray, s->nslp_data + finmsg.getLength(), GONE_CAPTOKENS*sizeof(unsigned int));
				
				for(int i = 0; i<GONE_CAPTOKENS; i++){
					GONEConnectionCapability cap = GONEConnectionCapability(s->mri->srcAddr.v4.addr,
																			s->mri->destAddr.v4.addr,
																			captokenarray[i]);
					capabilitytable->remove(&cap);
					//std::cerr<<"captokenarray["<< i <<"] = " << captokenarray[i] << std::endl;
				}
				
			}
			connections->remove(this_connection);
			delete this_connection;
			//std::cerr<< time(NULL) <<": Connection finished\n";
			struct timeval diezeit;
			gettimeofday(&diezeit, NULL);
			fprintf(stderr, "%i.%06d CONNECTION FINISHED\n", diezeit.tv_sec, diezeit.tv_usec);

			return;
		}
		else{
			GONE_VERBOSE(std::cerr<<" for someone else, length is " << s->nd_size << "\n";);
			
			//delete remaining capabilities
			if(s->nd_size-finmsg.captokenlist()*4 == finmsg.getLength()){
				
				unsigned int captokenarray[16];
				
				memcpy(captokenarray, s->nslp_data + finmsg.getLength(), GONE_CAPTOKENS*sizeof(unsigned int));
				
				for(int i = 0; i<GONE_CAPTOKENS; i++){
					GONEConnectionCapability cap = GONEConnectionCapability(s->mri->srcAddr.v4.addr,
																			s->mri->destAddr.v4.addr,
																			captokenarray[i]);
					capabilitytable->remove(&cap);
					//std::cerr<<"captokenarray["<< i <<"] = " << captokenarray[i] << std::endl;
				}
				
			}
			
			// pass fin msg. on
			api->SendMessage(s->nslp_data, s->nd_size, NULL, GONE_NSLP_ID,
							 (unsigned char *)s->sid,
							 (msg_routing_info_base_t *)s->mri, NULL, NULL, true, false, false,
							 2000, 10);
			return;
			
		}
		
	} // if(type == GONE_FIN)
	
	else{
		std::cerr<<"Got unknown packet\n";
	}
	
}


void GONEServer::handleMessageStatus(api_msg_status *s)
{
	printf("handleMessageStatus\n");
}

void GONEServer::handleNetworkNotification(api_nw_notification *s)
{
	printf("handleNetworkNotification\n");
}

int GONEServer::createServerSocket()
{
	printf("GONEServer::createServerSocket\n");
	
	int sock;
	struct sockaddr_un serv_addr;
	int len;
	
	// in case the app didn't exit cleanly
	unlink(NSLP_PATH_TO_UNIX_SOCKET);
	if ((sock = socket(AF_UNIX, SOCK_STREAM, 0)) < 0) {
		perror("GONE: socket creation failed");
		exit(-1);
	}
	
	memset((char *)&serv_addr, 0, sizeof(serv_addr));
	serv_addr.sun_family = AF_UNIX;
	strcpy(serv_addr.sun_path, NSLP_PATH_TO_UNIX_SOCKET);
	len = sizeof(serv_addr);
	
	if (bind(sock, (struct sockaddr *)&serv_addr,len) < 0) {
		perror("GONE: unable to bind unix socket");
		exit(-1);
	}
	
	// set bits to rwxrwxrwx
	if (chmod(NSLP_PATH_TO_UNIX_SOCKET, S_IRWXU|S_IRWXG|S_IRWXO) < 0) {
		perror("GONE: chmod on socket failed");
		exit(-1);
	}
	
	listen(sock, 5);
	return sock;
}

void GONEServer::closeSocket(int sock)
{
	api->removeSocket(sock);
	clientSockCount--;
}

void GONEServer::acceptClient(int sockfd, void * sockData)
{
	int l = 0;
	int sock = 0;
	struct sockaddr_un client_addr;
	
	memset(&client_addr, 0, l = sizeof(client_addr));
	if ((sock = accept(nslpSock, (struct sockaddr *) &client_addr, (socklen_t *) &l)) < 0) {
		perror("GONE: accept failed");
		exit(-1);
	}
	
	printf("GONE client connected\n");
	addSocket(sock);
}

void GONEServer::addSocket(int sock)
{
	clientSockCount++;
	api->addSocket(sock, NULL, clientCallback);
}
